;;;; Copyright (C) 1998 R. Dennis Cook and Sanford Weisberg.  Arc is 
;;;; XLISP-STAT computer code to accompany the book "An Introduction to
;;;; Regression Graphics", published by John Wiley, Inc.
;;;; All rights reserved.

;
;  various miscellaneous functions 
;

(defun sweep (a margin f &rest args)
"Args: a margin  f (&rest args)
Equivalent to the sweep function in S.  Takes 2 two-dimensional
array A, and produces an array of the same size and shape with f
applied to each column of A (margin = 'cols) or each row of A
(margin = 'rows)."
(case margin
  ('cols (apply #'bind-columns (mapcar f (column-list a))))
  ('rows (apply #'bind-rows (mapcar f (row-list a))))))

(defun substr (string start &optional (length (length string)))
"Args: (string start &optional (length (length string)))
Returns part of a string, starting at the start-th place, and extending
to length places."
  (subseq string start (min (+ start length) (length string))))

(defun which-not (a) 
"Function args: (a)
Returns indices of the elements of a list that equal nil."
  (which (mapcar #'not a)))

(defun backsolve (x y)
"Args:  (x  y)
Solves the triangular system xb=y for b, where x is an upper triangular 
matrix."
  (let* ((n (length y))
         (d (diagonal x))
         (x (row-list (- (sweep x 'cols #'(lambda (a) (/ a d))) 
                         (identity-matrix n))))
         (b (/ y d)))
   (dotimes (j (1- n))
            (let* ((nj2 (- n j 2)))
              (setf (select b nj2) (- (select b nj2) 
                   (inner-product (select x nj2) b)))))
    b))

(defun range (x) 
"Function args: (x)
Returns the minimum and maximum of x, ignoring missing values."
  (let ((z (select x (find-obs x)))) (list (min z) (max z))))

(defun midrange (x) 
"Function args: (x)
Returns the midrange of x, with missing values deleted."
  (* .5 (apply #'+ (range x))))

(defun cosangle (a B)
"Function args:  (a B)
Returns the cosine of the angle between a vector a and a subspace
spanned by B."
  (let* ((p (project b :intercept nil)))
    (cosangle1 a (send p :project a))))

(defun cosangle1 (a b)
"Function args:  (a b)
Returns the cosine of the angle between two lists or vectors a and b"
  (/ (inner-product a b) (sqrt (* (inner-product a a) (inner-product b b)))))

(defun angle (a b)
"Function args: (a B)
Returns the angle between vector a and subspace b in degrees"
  (* 180 (acos (cosangle a b)) (/ pi)))

;;; geometric mean

(defun geometric-mean (x &optional sel weights)
"Args:  (x &optional (sel (find-obs x)))
Returns the geometric mean if the elements of (select x sel)
are strictly positive, and returns nil otherwise."
 (let* ((sel (if sel sel (find-obs x)))
        (w (if weights weights (repeat 1 (length sel))))
        (z (cond (sel (select x sel)) (t x))))
  (if (> (min z) 0) (exp (/ (sum (* (log z) w)) (sum w))) nil)))

; helpful functions inherited from nonlinear regression protos.

(defun select-row (mat row-indices)
"Function args: (mat row-indices)
Returns the chosen rows of the matrix mat"
  (let* ((p (array-dimension mat 1)))
    (select mat row-indices (iseq p))))

(defun select-col (mat col-indices)
"Function args: (mat col-indices)
Returns the chosen cols of the matrix mat"
  (let  ((p (array-dimension mat 0)))
    (select mat (iseq p) col-indices)))

(defun normalize (v)
"Function args: (v)
Returns the vector v normalized to have (Euclidean) length one."
   (let* ((n (length v))
     (ss (* n (mean (* v v)))))
     (/ v (sqrt ss))))

(defun getrow (i coefs cum dim)
"Function args: (i coefs cum dim)
Internal function used to get the upper triangular for the symetric matrix B"
  (append (repeat 0 i)
          (select coefs (iseq (select cum i) (+ (select cum i) (- dim 1 i))))))

(defun coerce2 (x type)
"Function args: (x type) Modified coerce to handle matrices like in 
Xlisp-Stat 2.1R3."
  (if (arrayp x) (array-to-nested-list x) (coerce x type)))

(defun select-unique (x &rest args)
"Function args: (x &rest args)
Returns the indices of unique elements of x.  Args are passed to both 
position and remove-duplicates."
  (mapcar #'(lambda (a) (apply #'position a x args))
     (apply #'remove-duplicates x args)))

;;;
;;; project prototype -- Uses the QR decomposition to do many useful
;;;                      regression calculations
;;;

(defproto project-proto 
  '(x q r basis included weights intercept tol)
  ()
  *object*
  "Orthogonal Projections")

(defun project (x  &key  
                   (intercept t) (weights nil wset) (pweights nil pset)
                   (included t iset)
                   (tol 10e-7))
"Args:  (x  &key  
            (intercept t) weights pweights
            (included t iset)
            (tol 10e-7))
Computes the projection on the column space of x (with a column of
1s appended if intercept is t), using weights for a weight vector, and
excluding rows with nil elements of the included vector.  The method of 
computing is the qr factorization."
  (let* ((x (cond
              ((matrixp x) x)
              ((vectorp x) (bind-columns (coerce x 'list)))
              ((and (consp x) (consp (first x))) (apply #'bind-columns x))
              (t (bind-columns x)))) 
         (n (array-dimension x 0))
         (weights (cond
                   ((and pset (= (length pweights) n)) pweights)
                   ((and wset (= (length  weights) n)) weights)
                   (t (repeat 1 n))))
         (included (if iset included (repeat t n)))
         (h (send project-proto :new x intercept weights included tol)))
    h))

(defmeth project-proto :isnew (x  intercept weights included tol)
  (setf (slot-value 'intercept) intercept)
  (setf (slot-value 'weights) weights)
  (setf (slot-value 'included) included)
  (setf (slot-value 'tol) tol)
  (send self :x x))

(defmeth project-proto :compute () 
 (flet ((inverse1 (a) (if (equal (array-dimensions a) '(1 1))
                          (/ a) (inverse a))))
  (let* ((included (send self :included))
         (inc (which included))
         (not-inc (which (mapcar 'not included)))
         (sqrt-wt (sqrt (send self :weights)))
         (x (sweep (send self :x) 'cols #'(lambda (a) (* sqrt-wt a))))
         (intercept (send self :intercept))
         (size (array-dimensions x))
         (basis (if intercept (combine 0 (1+ (iseq (select size 1))))
                    (iseq (select size 1))))
         qr) 
    (when intercept (setf x (bind-columns sqrt-wt x)))
    (setf qr (send self :get-projection x inc basis)) 
    (setf (select x inc (first qr)) (second qr))
    (when not-inc
            (setf (select x not-inc (first qr)) 
                  (matmult (select x not-inc (first qr)) 
                           (inverse1 (third qr)))))
    (setf (slot-value 'q) (select x (iseq 0 (1- (nth 0 size))) (first qr)))
    (setf (slot-value 'r) (third qr))
    (setf (slot-value 'basis) 
          (if intercept (rest (- (first qr) 1)) (first qr)))
    (slot-value 'basis))))

(defmeth project-proto :get-projection (x inc basis)
"Method args: (x inc basis)
This method checks for zero pivots, and, if any are found, deletes them and
recursively calls itself."
  (let* ((qr (qr-decomp (select x inc basis)))
         (tol (send self :tol))
         (bad-pivot (position tol (abs (diagonal (second qr))) :test #'>)))
    (cond ((null bad-pivot) (cons basis qr))
          (t (send self :get-projection x inc 
                        (select basis (remove bad-pivot 
                                              (iseq (length basis))))))))) 

(defmeth project-proto :needs-computing (&optional set)
  (if set (setf (slot-value 'q) nil))
  (null (slot-value 'q)))

(defmeth project-proto :q (&optional firstp)
"Method args: (&optional firstp)
Computes the matrix consisting of the firstp columns of Q."
  (when (send self :needs-computing) (send self :compute))
  (let* ((q (slot-value 'q))
         (firstp (if firstp 
                  (+ (length (which (< (send self :basis) firstp)))
                     (if (send self :intercept) 1 0)) firstp)))
    (if firstp
        (select q (iseq 0 (1- (array-dimension q 0))) (iseq 0 (1- firstp)))
        q)))

(defmeth project-proto :q-inc (&optional firstp)
"Method args: (&optional firstp)
Returns the firstp columns of Q, but only the rows with inc = t."
 (let* ((q (send self :q firstp))
        (inc (which (send self :included))))
  (select q inc (iseq (array-dimension q 1)))))
  

(defmeth project-proto :r (&optional firstp)
"Method args: (&optional firstp)
Returns firstp columns of the upper-triangular matrix R."
  (when (send self :needs-computing) (send self :compute))
  (let* ((r (slot-value 'r))
         (firstp (if firstp 
                     (+ (length (which (< (send self :basis) firstp)))
                        (if (send self :intercept) 1 0)) firstp)))
      (if firstp (select r (iseq 0 (1- firstp)) (iseq 0 (1- firstp)))
      r)))

(defmeth project-proto :project (y &optional firstp)
"Method args: (y &optional firstp)
Computes QQ'y, where Q is the firstp columns of the QR factorization.  
Since QR is the factorization of W^{1/2}X, we want to compute
here W^{-1/2}QQ^TWY."
  (let* ((w (sqrt (send self :weights)))
         (q (copy-array (send self :q firstp)))
         (qvec (make-array (array-total-size q) :displaced-to q))
         (missing (find-missing (coerce qvec 'list)))
         (y (mapcar #'(lambda(y i) (if i y 0)) y (send self :included)))
         ans)
    (if missing (setf (select qvec missing) 0))
    (setf ans (/ (matmult q (matmult (transpose q) (* w y))) w))
    (if missing (setf (select ans (floor (/ missing (array-dimension q 1))))
              (set-missing)))
    ans))

(defmeth project-proto :rank (&optional firstp)
"Method args: ($optional firstp)
Returns the dimension of the space spanned by the firstp columns of X and by the constant column if intercept is t, or the dimension of R(1,X) by 
default."
   (let ((basis (send self :basis))
         (int (if (send self :intercept) 1 0)))
        (if firstp (+ int (length (which (if-else (< basis firstp) t nil))))
                   (+ int (length basis)))))

(defmeth project-proto :p2.1 (y firstp &optional secondp)
"Method args: (y firstp &optional secondp)
Computes (Q2Q2' - Q1Q1')y, where Q1 is the firstp columns and Q2 is the
secondp columns."
  (- (send self :project y secondp) (send self :project y firstp)))

(defmeth project-proto :residuals (y &optional (firstp nil))
"Method args: (y &optional (firstp nil))
Computes (I - QQ')y"
  (- y (send self :project y firstp)))

(defmeth project-proto :num-inc ()
"Method args: ()
Returns the number of included cases."
  (length (which (send self :included))))

(defmeth project-proto :residual-sum-of-squares (y &optional (firstp nil))
"Method args: (y &optional firstp)
Returns the resdiaul sum of squares for the regression of y on the firstp
columns of x."
  (sum (^ (select (send self :residuals y firstp) 
                  (which (send self :included))) 2)))

(defmeth project-proto :sigma-hat (y &optional (firstp nil))
"Method args: (y &optional firstp)
Returns sigmahat for the regression of y on the firstp
columns of x."
  (sqrt (/ (send self :residual-sum-of-squares y firstp)
           (- (send self :num-inc) (send self :rank firstp)))))

(defmeth project-proto :studentized-residuals (y &optional (firstp nil))
"Method args: (y &optional firstp)
Returns studentized residuals for the regression of y on the firstp
columns of x."
  (let* ((res (send self :residuals y firstp))
         (lev (send self :leverages firstp))
         (sig (send self :sigma-hat y firstp))
         (inc (send self :included)))
        (if-else inc
                 (/ res (* sig (sqrt (pmax .00001 (- 1 lev)))))
                 (/ res (* sig (sqrt (+ 1 lev)))))))

(defmeth project-proto :externally-studentized-residuals (y &optional firstp)
"Method args: (y &optional firstp)
Returns ext-studentized residuals for the regression of y on the firstp
columns of x."
  (let* ((res (send self :studentized-residuals y firstp))
         (np (- (send self :num-inc) (send self :rank firstp))))
      (if-else (send self :included)
               (* res (sqrt (/ (- np 1) (- np (^ res 2))))) res)))

(defmeth project-proto :qty (y &optional (firstp nil))
"Method args: (y &optional (firstp nil))
Computes Q'y, properly weighted, where Q is the firstp columns of Q."
  (let* ((w (sqrt (send self :weights)))
         (q (send self :q firstp))
         (inc (which (send self :included)))
         (b (iseq (array-dimension q 1))))
    (matmult (transpose (select q inc b)) (select y inc))))

(defmeth project-proto :coef-estimates (y &optional firstp)
"Method args: (y &optional (firstp nil))
Computes coefficient estimates by backsolving."
 (let ((z (if (send self :weights) (* (sqrt (send self :weights)) y) y)))
   (backsolve (send self :r firstp) (send self :qty z firstp))))

(defmeth project-proto :projection-matrix (&key firstp case-list)
"Method args: (&key firstp case-list)
Computes QQ' where Q has rows given by case-list and firstp columns."
  (let* ((q (send self :q firstp)))
    (when case-list (setf q (select q case-list 
                                    (iseq 0 (1- (array-dimension q 1))))))
    (matmult q (transpose q))))

(defmeth project-proto :leverages (&optional firstp)
"Method args: (&optional firstp)
Returns the diagonals of Q1Q1', where Q1 is firstp columns of Q."
  (let* ((q (send self :q firstp)))
    (mapcar #'(lambda (a) (sum (^ a 2))) (column-list (transpose q)))))

(defmeth project-proto :x (&optional new-x)
"Method args: (&optional new-x)
With no argument returns the x matrix as supplied to m. With an argument
NEW-X sets the x matrix to NEW-X and recomputes the estimates."
  (when new-x
        (setf (slot-value 'x)
               (cond
                 ((matrixp new-x) new-x)
                 ((vectorp new-x) (apply #'bind-columns (coerce new-x 'list)))
                 ((consp new-x) (apply #'bind-columns new-x))))
        (send self :needs-computing t))
   (slot-value 'x))

(defmeth project-proto :intercept (&rest args)
"Method args: (&optional new-intercept)
With no argument returns T if the model includes an intercept term, nil if
not. With an argument NEW-INTERCEPT the model is changed to include or exclude
an intercept, according to the value of NEW-INTERCEPT, and estimates are
recomputed."
  (when args 
        (setf (slot-value 'intercept) (car args))
        (send self :needs-computing t))
  (slot-value 'intercept))

(defmeth project-proto :tol (&optional tol)
"Method args: (&optional tol)
With no argument returns the tolerance."
  (when tol (setf (slot-value 'tol) tol))
  (slot-value 'tol))

(defmeth project-proto :weights (&optional new-w)
"Method args: (&optional new-w)
With no argument returns the weight sequence as supplied to m; NIL means
ans unweighted model. NEW-W sets the weights sequence to NEW-W and
recomputes the estimates."
  (when new-w (setf (slot-value 'weights) new-w) (send self :needs-computing t))
  (slot-value 'weights))

(defmeth project-proto :pweights (&optional (new nil set))
"Method args: (&optional (new nil set))
Sets/returns prior weights."
  (if set (send self :weights new) (send self :weights)))

(defmeth project-proto :basis ()
"Method args: ()
Returns the indices of the linearly independeent columns of X."
  (slot-value 'basis))

(defmeth project-proto :included (&optional new-included)
"Method args: (&optional new-included)
With no argument,  NIL means a case is not used in calculating
estimates, and non-nil means it is used.  NEW-INCLUDED is a sequence
of length of y of nil and t to select cases.  Estimates are recomputed."
  (when (and new-included 
             (= (length new-included) (length (slot-value 'included)))) 
        (setf (slot-value 'included) new-included) 
        (send self :needs-computing t))
  (slot-value 'included))

(defmeth project-proto :toggle-cases (&optional (idnum nil))
"Method args: (idnum), a list of case numbers
Deletes/restores specified cases.  If idnum is nil, all cases are restored."
(let* (included)
  (cond
    ((null idnum)
     (setf included (mapcar 'not (iseq 0 (1- (send self :num-cases))))))
    (t
     (setf included (send self :included))
     (dolist (i (coerce idnum 'list))
             (setf (nth i included) (not (nth i included))))))
  (send self :included included)))

(defmeth project-proto :num-cases ()
"Method args: ()
Returns the number of cases."
  (array-dimension (send self :q) 0))

;;;
;;;  Specialized transformation functions
;;;

(defun box-cox-power (y p &key included weights)
"Function args: (data power &key included)   
If y is strictly positive, this function returns the modified power
transformation of data.  See Atkinson, 1985, Plots, Transformations and
Regression, page 87, eq. 6.2.6."
  (when (<= (min y) 0)
        (message-dialog "Values must all be positive")
        (return-from box-cox-power))
  (let ((gm (geometric-mean y (if included (which included)) weights)))
       (cond ((< (abs p) .00001) (* gm (log y))) 
             (t (/ (- (^ y p) 1) (* p (^ gm (- p 1))))))))

(defun folded-power (y p &key included)
"Function args: (data power &key included)   
If y is strictly bounded in 0,1, this function returns the normalized
folded power transformation of data.  See Cook and Weisberg, 1982, 
Residuals and Influence in Regression, eq. 2.4.11."
 (let ((sel (if included (which included) (iseq (length y)))))
  (when (or (<= (min (select y sel)) 0) (>= (max (select y sel)) 1))
        (message-dialog "Values must all be strictly between 0 and 1")
        (return-from folded-power))
  (let ((1-y (- 1 y))
        (p-1 (- p 1))) 
       (cond ((> (abs p) .001)
              (/ (- (^ y p) (^ 1-y p))
                 (* p (geometric-mean (+ (^ y p-1) (^ 1-y p-1)) sel))))
             (t (* (geometric-mean (* y 1-y) sel) (log (/ y 1-y))))))))

(defun modulus-transform (y p &key included)
"Function args: (data power &key included)   
This function returns the normalized modulus
transformation of data.  See Cook and Weisberg, 1982, Residuals and Influence
in Regression, eq. 2.4.10."
  (let* ((sign (if-else (< y 0) -1 1))
         (ay1 (+ 1 (abs y)))
         (gm (geometric-mean ay1 (if included (which included)))))
       (cond ((< (abs p) .001) 
              (* sign gm (log ay1)))
             (t 
              (/ (* sign (- (^ ay1 p) 1))
                 (* p (^ gm (- p 1))))))))

(defun power-constructed-variable (y &key (power 1) included)
"Args:  (y &key (power 1) included)
Returns the score vector for testing Lambda = p in Box-Cox power 
transformation of y.  Reference:  Atkinson (1985), Plots, Transformations and
Regression, page 99."
  (when (<= (min y ) 0)
        (message-dialog "Values must all be positive")
        (return-from power-constructed-variable))
  (let ((gm (geometric-mean y (if included (which included))))
        (p power))
       (cond ((< (abs p) .0001)
              (* gm (log y) (- (* .5 (log y)) (log gm))))
             (t (- (/ (* (^ y p) (log y)) (* p (^ gm (- p 1))))
                   (/ (* (- (^ y p) 1) (+ (/ p) (log gm)))
                      (* p (^ gm (- p 1)))))))))

(defun scale-constructed-variable (y &key (shift 0) (power 1) included)
"Args:  (y &key (shift 0) (power 1) included)
Returns the score vector for testing shift = c in Box-Cox power 
transformation of (y - shift).   Reference:  Atkinson (1985), 
Plots, Transformations and Regression, page 185, eq 9.1.3."
  (when (<= (min (+ y shift)) 0)
        (message-dialog "Values must all be positive")
        (return-from scale-constructed-variable))
  (let* ((ys (cond (included (select y (which included))) (t y)))
         (gm (geometric-mean (+ ys shift) (if included (which included))))
         (hminv (/  (sum (/ 1 ys)) (length ys)))
         (ws (box-cox-power (+ y shift) power :included included)))
       (- (^ (/ (+ y shift) gm) (- power 1)) (* (- power 1) ws hminv))))

(defun truncate-at-val (x cut &key (test 'lessorequal) (value cut))
"Args:  (x cut &key (test #'<=) (value 0))
Returns x with all elements of x that pass the test (test x cut) set to value.
Using the default, all values of x less than or equal to cut will be set to
cut. Test can be either one of the words 'less 'lessorequal 'greater
'greaterorequal, or any function that evalates to true or nil."
 (let* ((f (case test
             ('less #'<) ('lessorequal #'<=) 
             ('greater #'>) ('greaterorequal #'>=)
             (t test))))
   (if-else (funcall f x cut) value x)))

(defun slice (&rest args) 
"Function args: (x cutpoints &key values)
Returns a list of the same length as x by discretizing x into categories.
Cutpoints can be a list of category boundaries, or an integer.   If an
integer, then then the range of x is divided into cutpoints equally spaced
quantiles.  If cutpoints is a list, then 1+(length cutpoints) categories
are created.  The keyword :values, if present gives the labels for the 
categories.  If values is a function like #'mean, then the label for each
group is the within-group mean."
  (apply #'cut args))

(defun cut (x cutpoints &key values)
"Function args: (x cutpoints &key values)
Returns a list of the same length as x by discretizing x into categories.
Cutpoints can be a list of category boundaries, or an integer.   If an
integer, then then the range of x is divided into cutpoints equally spaced
quantiles.  If cutpoints is a list, then 1+(length cutpoints) categories
are created.  The keyword :values, if present gives the labels for the 
categories.  If values is a function like #'mean, then the label for each
group is the within-group mean."
 (let* ((s (find-obs x))
        (a (make-list (length x) :initial-element (set-missing))))
  (setf (select a s) (cut1 (select x s) cutpoints :values values))
  a))

(defun cut1 (x cutpoints &key values)
 (let* ((c (combine (if (integerp cutpoints)
                        (quantile x (/ (iseq 1 (1- cutpoints)) cutpoints))
                        cutpoints) (/ machine-epsilon)))
        (z (mapcar #'(lambda (x) (position x c :test #'<=)) x)))
   (cond ((null values) (select (iseq (length c)) z))
         ((functionp values) 
          (select 
           (mapcar #'(lambda (j) 
            (let ((sel (select x (which (if-else (= z j) t nil)))))
              (if sel (funcall values sel)))) (iseq (length c))) z))
         (t (select values z)))))

(defun recode-value (x old new)
"Function args: (x old new)
Returns x with all occurances of old replaced by new."
 (cond ((listp old) (recode-values x old new))
       (t
  (let ((sel (which (mapcar #'(lambda (x) (equal x old)) x)))
        (z (copy-seq x)))
   (setf (select z sel) new)
   z))))

(defun recode-values (x old &optional (new (iseq (length old))))
"Function args: (x old new)
Returns x with all occurances of the jth elt of old 
replaced by the jth elt of new."
 (let* ((sel (mapcar #'(lambda (old)
               (which (mapcar #'(lambda (x) (equal x old)) x))) old))
        (z (copy-seq x)))
   (dolist (j (iseq (length old)))
     (setf (select z (select sel j)) (select new j)))
   z))

(defun qsort (x &rest args)
"Function args (x)
Alphabetizes the elements of x, even if they are numbers, strings or symbols,
with numbers first."
 (let ((x (copy-seq x)))
 (flet ((qcvx (a b)
     (cond ((and (numberp a) (numberp b)) (< a b))
           ((numberp a) t)
           ((numberp b) nil)
           (t (string-lessp (string-upcase a) (string-upcase b))))))
  (sort x #'qcvx))))

(defun recode (x &rest args)
"Function args: (x &key (print t))
Recodes the unique values of x to be integers, starting with 0.  If print is
t, a table describing the recoding is printed in the text window."
 (let* ((s (find-obs x))
        (a (make-list (length x) :initial-element (set-missing))))
  (setf (select a s) (apply #'recode1 (select x s) args))
  a))

(defun recode1 (x &key (print t) (alphabetize *alphabetize-labels*))
"Function args: (x)
Returns a list like x with the p unique elements of x replaced by the integers
0 to p-1."
   (let* ((labels (remove-duplicates x 
                    :test (if (every #'numberp x) #'= #'equal)))
          (old (cond ((null (position nil (mapcar #'numberp labels))) 
                      (qsort labels))
                     (alphabetize (qsort labels))
                     (t labels)))
;         (old (qsort (reverse (remove-duplicates x :test #'equal))))
          (sel (mapcar #'(lambda (a) (position a old :test #'equal)) x)))
    (when print
      (format t "Recoded variable:~%")
      (format t "Original value  Recoded as~%")
      (format t "--------------  ----------~%")
     (mapcar #'(lambda (old j)
      (format t "~12a  ~17t~a~%" old j)) old (iseq (length old))))
    (select (iseq (length old)) sel)))

;;;
;;; Case indicator function
;;;
(defun case-indicator (selection length)
"Function args: (selection args)
Returns a list of length length with all zeros, except of ones in the elements
given by the selection."
 (let* ((ans (make-list length :initial-element 0)))
  (setf (select ans selection) 1)
  ans))

(defun jitter (x &optional 
   (by (* *jitter-radius* (standard-deviation (remove-missing x)))))
"Function args: (x &optional (by (* *jitter-radius* (standard-deviation x))))
Returns a \"jittered\" version of x by adding a uniform random number 
with range by to x."
 (+ x (- (* by (uniform-rand (length x))) (/ by 2))))

(defun read-list-columns (data p)
"Function args:  data p
Equivalent to the read-data-columns function, except that data is a list of
numbers."
   (transpose (split-list data p)))

(defun outlier-pvalue (tval df n)
"Function args:  (tval df n)
Uses the Bonferroni inequality to approximate the p-value of the t-test that
the largest externally studentized residual corresponds to an outlying case."
  (min 1 (* n (* 2 (- 1 (t-cdf (abs tval) df))))))

(defun t-pvalue (z &optional (df 10000) (what 'twotail))
  (case what
        (twotail (* 2 (- 1 (t-cdf (abs z) df))))
        (uppertail (- 1 (t-cdf z df)))
        (lowertail (t-cdf z df))))

;;;
;;;  Golden Section Search algorithm
;;;
(defun golden (g a b &key (epsilon .01) (min t))
"Function args: (g a b &key (epsilon .01) (min nil) (plot nil))
Golden section search.  Algorithm 4.5.1 from R. Thisted (1988), Elements of 
Statistical Computing, Chapman-Hall, p. 203.  Finds the maximum of a function g 
of a single parameter unless :min is t; then, it finds the minimum." 
  (let* ((alpha (/ (- 3 (sqrt 5)) 2))
         (f (if min #'(lambda (a) (- (funcall g a)))
                    #'(lambda (a)    (funcall g a))))
         (L a)
         (U b)
         (W (- U L))
         (t1 (+ L (* 1 alpha W)))
         (t2 (- U (* 1 alpha W)))
         (f1 (funcall f t1))
         (f2 (funcall f t2))
         (vals (list t1 t2))
         (crit (list f1 f2)))
        (do () ((< (abs (- t1 t2)) epsilon))
            (cond ((< f1 f2)
                   (setf L t1)
                   (setf t1 t2)
                   (setf f1 f2)
                   (setf W (- U L))
                   (setf t2 (- U (* alpha W)))
                   (setf f2 (funcall f t2))
                   (setf vals (cons t2 vals))
                   (setf crit (cons f2 crit)))
                  (t
                   (setf U t2)
                   (setf t2 t1)
                   (setf f2 f1)
                   (setf W (- U L))
                   (setf t1 (+ L (* alpha W)))
                   (setf f1 (funcall f t1))
                   (setf vals (cons t1 vals))
                   (setf crit (cons f1 crit)))))
        (when min (setf crit (- crit)))
         (list (* .5 (+ t1 t2)) vals crit (* (if min -.5 .5) (+ f1 f2)))))

(defun gl (a b &optional (n (* a b)))
"Function args: (a b &optional (n (* a b)))
Generate a list of length n of the numbers 0 to a-1 in groups of size b.
For example (gl 2 3 12) will give 0 0 0 1 1 1 0 0 0 1 1 1. Similar to a GLIM
feature of the same name."
 (let ((out (repeat (iseq a) (repeat b a))))
      (do ((c out (combine c c)))
          ((<= n (length c)) (select c (iseq n))))))

(defun datacolumns (n &rest data) (transpose (split-list data n)))
(defun datarows (n &rest data) (split-list data n))

(defun not-constant (x &key (test #'=))
"Function args: (x)
Returns nil if x is a list with all elements =, and t otherwise"
  (> (apply #'- (reverse (range x))) (* 2 machine-epsilon)))

(defun remove-constant-columns (lists)
"Function args: (lists)
Returns the list of indices of columns that are not constant."
  (which (mapcar #'not-constant lists)))

(defmeth edit-text-item-proto :value (&key (blank 0))
"Method args: ()
Edit-text-item-proto normally returns a string.  This method converts
the string to a number, if possible."
  (if (equal (send self :text) "") blank
      (with-input-from-string (s (send self :text)) (read s))))

(defmeth edit-text-item-proto :evaluate ()
  (let* ((text (send self :text)))
   (evaluate-string-to-values text)))

(defun evaluate-string-to-values (text)
"Function args: (text)
Uses parcil to convert text to a function, and then the function is
evaluated."
 (let* ((string (string-trim " " text))
        (expression (read (make-string-input-stream string)))
        (parens (and  (equal #\( (select string 0))  
                      (equal #\) (select (reverse string) 0)))))
  (cond ((null parens) (eval (parcil string)))
        ((is-lisp expression) (eval expression))
        (t  (eval (parcil string))))))

(defun is-lisp (expr)
 (let* ((x (first expr))
        (ans (ignore-errors (eval `(functionp #',x)))))
   ans))

(defun string-to-number (val)
 (with-input-from-string (s val) (read s)))

;;;
;;; two list proto types -used in lots of dialogs
;;;

(defun twolist-dialog (names &rest args &key 
         (left-title "Candidates") (left-initial names)
         (right-title "Selection") (right-initial (repeat "" (length names)))
         (title "Make selection...")
         help)
  (let* ((all (send allocate-items-proto :new))
         (left (send subordinate-list-proto :new left-initial 'primary all))
         (lt (send text-item-proto :new left-title))
         (right (send subordinate-list-proto :new right-initial
                      'secondary all))
         (rt (send text-item-proto :new right-title))
         (title (send text-item-proto :new title))
         (cancel (send modal-button-proto :new "Cancel"))
         (ok (send modal-button-proto :new "OK" :action 
                   #'(lambda () (send all :value)))) 
         (helpb (send modal-button-proto :new "Help" :action
                      #'(lambda () (apply #'twolist-help help names args))))
         (dialog (send modal-dialog-proto :new
                       (list 
                        (list title)
                        (list 
                         (list left-title left)
                         (list right-title right))
                        (if help (list ok cancel helpb)
                                        (list ok cancel))))))
    (send dialog :default-button ok)
    (send dialog :modal-dialog)))

(defun twolist-help (string &rest args)
  (message-dialog string)
  (apply #'twolist-dialog args))
  

(defproto allocate-items-proto 
  '(actors hilight-item hilight-actor) () *object*)

(defmeth allocate-items-proto :add-actor (actor)
   (setf (slot-value 'actors) (append (slot-value 'actors) (list actor)))) 

(defmeth allocate-items-proto :do-action (item actor double)
  (let ((hilight-item (slot-value 'hilight-item))
        (hilight-actor (slot-value 'hilight-actor))
        (actor-type (send actor :slot-value 'type))
        (single (null double))
        (blank (equal item "")))  
  (cond ((and single (null blank) (null hilight-item))
         (setf (slot-value 'hilight-item) item)
         (setf (slot-value 'hilight-actor) actor))
        ((and single (null blank) (eq actor hilight-actor))
         (setf (slot-value 'hilight-item) item)
         (setf (slot-value 'hilight-actor) actor))
        ((and single hilight-item)
         (unless (eq actor hilight-actor) (send self :clear-selection))
         (send self :update hilight-item :from hilight-actor :to actor))
        ((and single blank (null hilight-item))
         (send self :clear))
        ((and double (null blank) (eq actor-type 'primary))
         (send self :update-to-secondary item :from actor))
        ((and double (null blank))
         (send self :update item :from actor))
        ((and double blank)
         (send self :clear-selection)
         (send self :clear-hilight)))))

(defmeth allocate-items-proto :clear-selection ()
  (mapcar #'(lambda (a) (send a :selection nil)) (slot-value 'actors)))
  
(defmeth allocate-items-proto :clear-hilight ()
  (setf (slot-value 'hilight-item) nil)
  (setf (slot-value 'hilight-actor) nil))
    
(defmeth allocate-items-proto :clear ()
   (send self :clear-selection)
   (send self :clear-hilight))
     
(defmeth allocate-items-proto :primary-actor ()
  (let* ((actors (slot-value 'actors)))
    (first (select actors (which 
        (mapcar #'(lambda (a) (eq (send a :slot-value 'type) 'primary))
                    actors))))))

(defmeth allocate-items-proto :secondary-actor ()
  (let* ((actors (slot-value 'actors)))
    (select actors (which 
        (mapcar #'(lambda (a) (eq (send a :slot-value 'type) 'secondary))
                    actors)))))

(defmeth allocate-items-proto :update-to-secondary (item &key from)
  (let* ((a (send self :secondary-actor))
         (blank (mapcar #'(lambda (a) (send a :add item nil)) a))
         (j (position t blank)))
        (when j (send self :update item :from from :to (select a j)))))

(defmeth allocate-items-proto :update (item &key from 
    (to (send self :primary-actor)))
  (let* ((from-type (send from :slot-value 'type))
         (to-type (send to :slot-value 'type))) 
    (cond ((and (eq from-type 'primary) (eq to-type 'secondary))
           (when (send to :add item) (send from :delete item)))
          ((and (eq from-type 'primary) (eq to-type 'oneitem))
           (when (send to :add item) (send from :delete item)))
          ((and (eq from-type 'primary) (eq to-type 'duplicate))
           (send to :add item))
          ((and (eq from-type 'secondary) (eq to-type 'primary))
           (when (send to :add item) (send from :delete item)))
          ((and (eq from-type 'oneitem) (eq to-type 'primary))
           (when (send to :add item) (send from :delete item)))
          ((and (eq from-type 'duplicate) (eq to-type 'primary))
           (send from :delete item))
          ((eq from to) 
           (setf (slot-value 'hilight-item) item)
           (setf (slot-value 'hilight-actor) from)
           (return-from :update))
          ((null (eq to-type 'primary)) (send self :clear))
          (t (return-from :update)))
  (send self :clear)))

(defmeth allocate-items-proto :value ()
  (mapcar #'(lambda (a) (send a :value)) (slot-value 'actors)))

(defmeth allocate-items-proto :all-values ()
  (mapcar #'(lambda (a) (send a :all-values)) (slot-value 'actors)))

(defproto subordinate-list-proto '(type owner items) () list-item-proto)

(defmeth subordinate-list-proto :isnew (items type owner)
   (call-next-method items)
   (setf (slot-value 'type) type)
   (setf (slot-value 'owner) owner)
   (send owner :add-actor self)
   self)

(defmeth subordinate-list-proto :do-action (&optional double)
  (let* ((index (send self :selection))
         (item (select (coerce (slot-value 'list-data) 'list) index))
         (owner (slot-value 'owner)))
        (send owner :do-action item self double)))

(defmeth subordinate-list-proto :add (item &optional (update t))
  (let* ((j (position "" (coerce (slot-value 'list-data) 'list) 
                      :test #'equal)))
    (if (and j update) (send self :set-text j item))
    (if j t nil)))

(defmeth subordinate-list-proto :delete (item)
  (let* ((items (coerce (slot-value 'list-data) 'list))
         (p (length items))
         (j (position item items :test #'equal)))
    (if (< j (1- p))
        (mapcar #'(lambda (k) (send self :set-text k (select items (1+ k)))) 
            (iseq j (- (length items) 2))))
    (send self :set-text (1- p) "")))

(defmeth subordinate-list-proto :value ()
  (remove "" (coerce (slot-value 'list-data) 'list) :test #'equal))

(defmeth subordinate-list-proto :all-values ()
  (coerce (slot-value 'list-data) 'list))

(defun fill-list (items len &optional (char ""))
 (let* ((ans (repeat char len)))
    (setf (select ans (iseq (length items))) items)
    ans))
;;;
;;; A replacement for message dialog to allow for longer strings
;;; (the mac allows only 255 characters).
;;;
(defun big-message-dialog (&rest args)
"Function args: (&rest args)
Creates a message dialog by appending the args, one below the next."
  (let* ((items (mapcar #'(lambda (a) (send text-item-proto :new a)) args))
         (ok (send modal-button-proto :new "OK"))
         (dialog (send modal-dialog-proto :new 
                     (append (mapcar #'list items) (list ok)))))
     (send dialog :modal-dialog)))

(defun prnum (x)
"Function args: (x)
If x is nil,  returns \"NIL\".  If x is an integer, returns \"x\".  If x is
another number, returns x rounded to four decimals in quotes."
  (cond ((integerp x) (format nil "~d" x))
        ((and (numberp x) (< (abs (- x 0)) .00005) "0"))
        (t (string-trim " 0" (format nil "~7,4f" x)))))

(defun fun-print (function)
"Method args: (function)
Pretty prints the function"
  (pprint (function-lambda-expression function)))

(defun meth-print (object method)
"Method args: (object method)
Pretty prints the method for the object"
  (fun-print (send object :get-method method)))

;;
;;  Functions for missing data evaluation and handling
;;

(defun is-missing (a) 
  (or (complexp a) (and (realp a) (null (> a negative-infinity)))))

;;;
;;; Checks to see if 'not-a-number is bound.  If not, a message is printed,
;;; and the missing value code is set to '?.  Many computations with missing
;;; values will fail.
;;;
(defun set-missing ()
 (case *computer-type*
  (unix (cond ((boundp 'not-a-number) not-a-number) (t (not-ieee))))  
  (macintosh not-a-number)
  (windows not-a-number)))

(defun not-ieee ()
 (format t " 
; WARNING---This machine is not IEEE compliant.  Use of missing values
;           may cause errors.~%") '?)

(defun find-missing (a) 
"Function args: (a)
Returns the indices of the elements of a that are missing value
codes."
(which (mapcar #'(lambda (a) (is-missing a)) (coerce a 'list))))

(defun find-obs (x) 
"Function args: (x)
Returns the indices of the elements of x that are not missing value
codes."
  (reverse (set-difference (iseq (length x)) (find-missing x))))

(defun funcall-with-missing (f x)
"Function args (f x)
Funcalls f to the observed part of x."
 (funcall f (select x (find-obs x))))

(defun with-missing (f x)
"Function args: (f x)
Returns a list of length x with missing values unchanged, and with
f applied to remaining elements."
 (let* ((sel (find-obs x))
        (n (funcall f (select x sel)))
        (o (copy-seq x)))
  (setf (select o sel) n)
  o))

(defun missing-to-symbol (a &optional (value '??)) 
"Function args: (x &optional (value '??))
Returns an object like x with missing values replaced by value."
    (mapcar #'(lambda (a) (if (is-missing a) value a)) (coerce a 'list)))

(defun remove-missing (a) 
"Function args: (a)
Returns a list like a with missing value codes removed."
(if a
  (select a (reverse (set-difference (iseq (length a)) (find-missing a))))))

;;;
;;; critical-value calculator
;;; 5 August 1997; revised from original by I. Pardoe
;;;

(defun quantile-calculator ()
"Function args: ()
Opens a dialog to compute quantiles."
 (let ((ans (critical-value-dialog)))
  (when ans (apply #'print-quantile ans))))

(defun critical-value-dialog (&rest args &key (value "") (distribution 0) 
(deg1 "") (deg2 "") )
"Function args: ()
Opens a dialog to compute a critical value."
  (let* ((title (send text-item-proto :new "Calculate quantile..."))
	      (sigtext (send text-item-proto :new
"Computes Q(distribution, fraction).  Example:  Q(Normal,.75) 
is a value c such that Pr(z<c)=.75, with z a standard normal
random variable.  In this dialog, select the distribution you 
want and the value of the fraction."))
         (val-l (send text-item-proto :new "Fraction:"))
         (val   (send edit-text-item-proto :new value 
                       :text-length (apply #'ms *text-width-min*)))
         (dist-l (send text-item-proto :new "Distribution:"))
         (dist   (send choice-item-proto :new
                      (list "Normal" "t" "Chi^2" "F") 
                       :value distribution))
         (df-l (send text-item-proto :new "Degrees of Freedom:"))
         (df1 (send edit-text-item-proto :new deg1 
                      :text-length (apply #'ms *text-width-min*)))
         (df2 (send edit-text-item-proto :new deg2 
                      :text-length (apply #'ms *text-width-min*)))
         (typetext (send text-item-proto :new
"The normal distribution does not require degrees of freedom.
t and Chi^2 require choosing d.f.
F requires choosing both numerator and denominator d.f."))
         (cancel (send modal-button-proto :new "Cancel"))
         (get-answer #'(lambda () 
            (let* ((x (send val :value))
                   (f (send dist :value))
                   (d1 (send df1 :value))
                   (d2 (send df2 :value))
                   (tail 2))
               (cond ((equal (send val :text) "")
                      (message-dialog "Fraction not given.")
                      (critical-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)))
                     ((or (<= x 0) (>= x 1))
                      (message-dialog "Fraction not between 0 and 1.")
                      (critical-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)))
                     ((and (= f 1) (<= d1 0))
                      (message-dialog "D.F. for t not given.")
                      (critical-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)))
                     ((and (= f 2) (<= d1 0))
                      (message-dialog "D.F. for Chi^2 not given.")
                      (critical-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)))
                     ((and (= f 3) (or (<= d1 0) (<= d2 0)))
                      (message-dialog "Values of d.f. for F not given.")
                      (critical-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)))
                     (t (list :value x
                              :distribution
                                 (select (list 'normal 'tdist 'chisq 'f) f)
                              :df1 d1 :df2 d2))))))
         (done (send modal-button-proto :new "OK" :action 
                     #'(lambda () (funcall get-answer))))
         (help (send modal-button-proto :new "Help" :action
                     #'(lambda () (quantile-help args))))
         (dialog (send modal-dialog-proto :new
                   (list (list title) (list sigtext) (list typetext)
                         (list (list dist-l dist) (list df-l df1 df2)
                               (list val-l val))
                         (list done cancel help)))))
        (send dialog :default-button done)
        (send dialog :modal-dialog)))

(defun quantile-help (args)
	(big-message-dialog 
"Calculate quantiles for four distributions, Normal, t, 
Chi-squared, and F."
"Input a fraction f, a number between 0 and 1.  Choose 
a distribution d by selecting the name of the
distribution you want.  If the distribution is not the
normal, you must also specify degrees of freedom, one
value for t or Chi-squared, and two values for F."
"The value returned is c = Q(d,f), which is a solution
to the equation Pr(z < c) = f, where z is a random variable
with distribution d.")
(apply #'critical-value-dialog args))
	
(defun print-quantile (&key value distribution df1 df2 (verbose t))
 (flet ((f (v)
         (case distribution
          ('normal (normal-quant v))
          ('tdist (t-quant v df1))
          ('chisq (chisq-quant v df1))
          ('f (f-quant v df1 df2)))))
  (let ((name (case distribution
               ('normal "Normal")
               ('tdist (format nil "t_~a" df1))
               ('chisq (format nil "Chi-squared_~a" df1))
               ('f     (format nil "F_(~a,~a)" df1 df2)))))
      (if verbose (format t 
"Quantile calculation:  Q(~a,~a) = ~a~%" name value (f value))
                  (f value)))))

;;;
;;; p-value calculator
;;; 1 August 1997
;;;

(defun p-value-calculator ()
 (let ((ans (p-value-dialog)))
  (when ans (apply #'print-p-value ans))))

(defun p-value-dialog (&rest args &key (value "") (distribution 0) (deg1 "") 
(deg2 "") (tail 0))
"Function args: ()
Opens a dialog to compute a probability for a standard distribution."
  (let* ((title (send text-item-proto :new "Probability calculator"))
         (val-l (send text-item-proto :new "Value of statistic"))
         (val   (send edit-text-item-proto :new value 
                    :text-length (apply #'ms *text-width-short*)))
         (dist-l (send text-item-proto :new "Distribution:"))
         (dist   (send choice-item-proto :new
                      (list "Normal" "t" "Chi^2" "F") 
                       :value distribution))
         (dftext (send text-item-proto :new
"For t and Chi^2, you must enter one value 
for D. F.  For F, you must enter two values."))
         (df-l (send text-item-proto :new "D. F. ="))
         (df1 (send edit-text-item-proto :new deg1 
                   :text-length (apply #'ms *text-width-min*)))
         (df2 (send edit-text-item-proto :new deg2 
                   :text-length (apply #'ms *text-width-min*)))
         (type-l (send text-item-proto :new "Choose type"))
         (type   (send choice-item-proto :new
                      (list "Upper tail" "Two tail (t and normal only)" 
                            "Lower tail")
                      :value tail))
         (cancel (send modal-button-proto :new "Cancel"))
         (get-answer #'(lambda () 
            (let* ((x (send val :value))
                   (f (send dist :value))
                   (d1 (send df1 :value))
                   (d2 (send df2 :value))
                   (tail (send type :value)))
               (cond ((equal (send val :text) "")
                      (message-dialog "Value of statistic not given.")
                      (p-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)
                       :tail (send type :value)))
                     ((and (= f 1) (<= d1 0))
                      (message-dialog "Value of df for t not given.")
                      (p-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)
                       :tail (send type :value)))
                     ((and (= f 2) (<= d1 0))
                      (message-dialog "Value of df for Chisq not given.")
                      (p-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)
                       :tail (send type :value)))
                     ((and (= f 2) (< x 0))
                      (message-dialog "Value of Chisq cannot be negative.")
                      (p-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)
                       :tail (send type :value)))
                     ((and (= f 3) (or (<= d1 0) (<= d2 0)))
                      (message-dialog "Values of df for F not given.")
                      (p-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)
                       :tail (send type :value)))
                     ((and (= f 3) (< x 0))
                      (message-dialog "Value of F cannot be negative.")
                      (p-value-dialog :value (send val :text)
                       :distribution (send dist :value)
                       :deg1 (send df1 :text) :deg2 (send df2 :text)
                       :tail (send type :value)))
                     (t (list :value x
                          :distribution 
                              (select (list 'normal 'tdist 'chisq 'f) f)
                          :df1 d1 :df2 d2
                          :tail (select (list 'upper 'two 'lower) tail)))))))
         (done (send modal-button-proto :new "OK" :action 
                     #'(lambda () (funcall get-answer))))
         (help (send modal-button-proto :new "Help" :action
                     #'(lambda () (pvalue-help args))))
         (dialog (send modal-dialog-proto :new
                   (list (list title)
                         (list val-l val)
                         (list dist-l dist) (list dftext)
                         (list df-l df1 df2)
                         (list type-l type)
                         (list done cancel help)))))
        (send dialog :default-button done)
        (send dialog :modal-dialog)))

(defun pvalue-help (args)
	(big-message-dialog 
"Calculate a probability for a Normal, t, Chi-squared, 
or F distribution."
"Input a value s of the statistic and its degrees of 
freedom, and the upper, lower, or two-tailed probability is 
returned."
"The upper-tailed probability is the proportion of the distribution 
to the right of s. The lower-tailed probability is the proportion 
to the left of s. For Normal and t, the two- tailed probability is 
twice the proportion to the right of |s|.")
(apply #'p-value-dialog args))

(defun print-p-value (&key value distribution df1 df2 tail (verbose t))
 (flet ((f (v)
         (case distribution
          ('normal (normal-cdf v))
          ('tdist (t-cdf v df1))
          ('chisq (chisq-cdf v df1))
          ('f (f-cdf v df1 df2)))))
  (let* ((pvalue (case tail
                    ('upper (- 1 (f value)))
                    ('lower (f value))
                    ('two (* 2 (f (- (abs value))))))))
  (when verbose (case distribution
   ('normal 
     (format t "Normal dist., value = ~a, ~a-tail probability = ~a~%"
       value (string-downcase tail) pvalue))
   ('tdist (format t 
    "t dist. with ~a df, value = ~a, ~a-tail probability = ~a~%"
       df1 value (string-downcase tail) pvalue))
   ('chisq (format t 
    "Chisq dist. with ~a df, value = ~a, ~a-tail probability = ~a~%"
       df1 value (string-downcase tail) pvalue))
   ('f (format t 
    "F dist. with (~a, ~a) df, value = ~a, ~a-tail probability = ~a~%"
       df1 df2 value (string-downcase tail) pvalue)))))))

(defun paste (&rest args)
"Function args: (&rest variables &key (sep \" \" set))
Concatenates lists of numbers or strings to create one list.  For example if
x is '(\"yes\" \"yes\" \"no\" \"no\") and y is '(0 1 0 1), then 
> (paste x y :sep \"-\")
(\"yes-0\" \"yes-1\" \"no-0\" \"no-1\")"
 (let* ((len (length args))
        (set (cond ((> len 2) (eq :sep (select args (- len 2))))
                   (t nil))))
 (apply #'paste1 (if set (first (last args)) " ")
                 (if set (select args (iseq (- len 2))) args))))

(defun paste1 (sep &rest args)
 (case (length args)
       (1 args)
       (2 (mapcar #'(lambda (a b) (format nil "~a~a~a" a sep b))
           (first args) (second args)))
       (t (apply #'paste1 sep (append (list (mapcar #'(lambda (a b) 
           (format nil "~a~a~a" a sep b))
           (first args) (second args))) (rest (rest args)))))))

;;; get-decoded-time changes month number with version 3.52.10  (12/1/1999)

(defun time-stamp ()
 (let* ((time ( multiple-value-list (get-decoded-time)))
        (weekdays (list "Sun" "Mon" "Tue" "Wed" "Thu" "Fri" "Sat" ))
        (offset (if (and (> xls-subminor-release 9)
                         (>= xls-minor-release 52)) 1 0))
        (months (list "Jan" "Feb" "Mar" "Apr" "May" "Jun" "Jul"
                      "Aug" "Sep" "Oct" "Nov" "Dec" )))
  (apply #'format nil "~a ~a ~a, ~a, ~d:~2,'0d:~2,'0d"
   (select weekdays (select time 6))
   (select months (- (select time 4) offset))
   (select time '(3 5 2 1 0)))))

;;; 1/20/97, needed for Windows only.  The handling of not-a-number
;;; is different in Windows than in Unix or Macintosh.  In particular:
;;;   (^ not-a-number 2)  = not-a-number on Unix and Mac
;;;                       = 0 on Windows
;;;   (/ not-a-number not-a-number) = not-a-number on Unix and Mac
;;;                                 = illegal zero arg on Windows
;;;   (log not-a-number) = not-a-number on Unix and Mac
;;;                      = illegal zero arg on Windows
;;; These overrides avoid taking an integer power and dividing on Windows
;;; only.  They have no effect on other systems.

;(defun div (a1 &optional a2)
;"Function args (a1 &optional a2)
;Divides the list a1 elementwise by the list a2, or, if no a2, return the
;inverses of a1.  Handles missing values correctly, even on Windows."
; (let* ((num (if a2 a1 (repeat 1 (length a1))))
;        (denom (if a2 a2 a1)))
;     (cond ((eq *computer-type* 'windows) 
;            (mapcar #'(lambda (n d) (if (is-missing d) not-a-number (/ n d)))
;                num denom))
;      (t (/ num denom)))))
;;; override written by Luca Scrucca
(defun div (a1 &optional a2)
"Function args (a1 &optional a2)
Divides the list a1 elementwise by the list a2, or, if no a2, return the
inverses of a1.  Handles missing values correctly, even on Windows."
 (let* ((num (if a2 a1 (repeat 1 (length a1))))
        (denom (if a2 a2 a1))
        (result nil))
     (cond ((eq *computer-type* 'windows) 
            (unless (listp a1)  (setf num   (list num))
                                (setf denom (list denom)))
            (setf result (mapcar #'(lambda (n d) 
                                   (if (is-missing d) not-a-number (/ n d)))
                          num denom))
            (if (listp a1) result (car result)))
      (t (/ num denom)))))

;;  March 2, 2000

(defun div1 (a1 &optional a2)
"Function args (a1 &optional a2)
Divides a1 by a2, or, if no a2, returns the inverse of a1.  
Handles missing values correctly, even on Windows."
 (let* ((n (if a2 a1 1))
        (d (if a2 a2 a1)))
     (cond ((eq *computer-type* 'windows) 
            (if (is-missing d) not-a-number (/ n d)))
      (t (/ n d)))))

(defun powr (x p)
(cond ((eq *computer-type* 'windows) 
       (mapcar #'(lambda (x) (if (is-missing x) 
                   (apply #'* (repeat x p)) (^ x p))) x))
      (t (^ x p))))

;;;
;;; leverage contours
;;;
(defun leverage-contours (x1 x2 h)
"Function args: (x1 x2 h)
Returns the contours of constant leverage = h for the 2D predictor
x1 x2."
   (let* ((c (mapcar #'mean (list x1 x2)))
          (m (inverse (cross-product (apply #'bind-columns (- (list x1 x2) c)))))
          (r (sqrt (- h (/ 1 (length x1))))))
    (ellipse m r c)))

(defmeth scatterplot-proto :leverage-contours (&key (h '(.02 .04 .08 .1)))
"Method args: (&key h)
Adds leverage contours to visible data."
   (let* ((sel (send self :points-showing))
          (x (send self :point-coordinate 0 sel))
          (y (send self :point-coordinate 1 sel)))
     (mapcar #'(lambda (h) (send self :add-lines (leverage-contours x y h))) h)))

(defun make-case-indicator (n i)
"Function args: (n i)
Returns a list of length n of all zeros, execpt a one in the ith place.
Remember that counting of places starts with zero, not one."
 (let ((a (repeat 0 n))) (setf (select a i) 1) a)) 


(defun lag (x &key (by 1) (circular nil))
"Function args: (x &key (by 1) (circular nil))
Returns a list like x by displaced by by units.  If circular is nil,
the first (or last) by places of the answer are set to missing; 
otherwise, the lag is circular."
 (let ((ans (copy-seq (select x (iseq (- (length x) by)))))
       (front (if circular (select x (iseq (- (length x) by) (1- (length x))))
                  (repeat (set-missing) by))))
  (append front ans)))

(defun rect-coords (r theta)
"Functions args: (r theta)
Converts the n polar coordinates r theta into rectangular coordinates x y."
 (list (* r (cos theta)) (* r (sin theta))))

(defun polar-coords (x y)
"Function args: (x y)
x and y are lists of length n.  Returns to lists r, theta, converting x y to
polar coordinates, theta in radians."
  (list (polar-radius x y) (polar-angle x y)))

(defun polar-angle (x y)
"Function args: (x y)
Returns the polar angles corresponding to the lists of rectangular coordinates
x, y."
 (if (and (listp x) (listp y)) (mapcar #'atan1 x y) (atan1 x y)))

(defun polar-radius (x y)
"Function args: (x y)
Returns the polar radius corresponding to the lists of rectangular coordinates
x, y."
 (sqrt (+ (^ x 2) (^ y 2))))

(defun atan1 (x y) 
"Function args: (x y)
Returns the arc tangent of (/ x y), with angle between 0 and 2pi."
 (let* ((a (cond ((= x 0) (/ pi 2)) (t (atan (abs (/ y x)))))))
  (cond ((and (>= x 0) (>= y 0)) a)
        ((and (< x 0)  (>= y 0)) (- pi a))
        ((and (< x 0)  (< y 0) ) (+ pi a))
        ((and (>= x 0) (< y 0) ) (- (* 2 pi) a)))))
  
;;
;; function to display help for all the methods for a prototype
;;
(defun display-help (object)
"Function args: (object)
Prints the documentation for all of the objects own methods."
   (format t "==========================================~%")
   (format t "~a precedence list:~%" (send object :slot-value 'proto-name))
   (mapcar #'(lambda (p) (format t "~a~%" (send p :slot-value 'proto-name))) 
                         (send object :precedence-list))
   (mapcar #'(lambda (m) 
             (format t "------------------------------------------~%")
             (format t "~a: " (send object :slot-value 'proto-name))
             (format t "~a~%" (send object :help m)))
   (qsort (remove-duplicates (send object :own-methods) :test #'equal)))
   t)

(defun truncate-string (string &key (loc (apply #'ms *text-width-max*))) 
  (substr string 0 loc))

; rewritten Sept 17, 1999 to split at blanks or math symbols if possible
(defun split-string (string &key (loc 60));(apply #'ms *text-width-max*)))
"Function args:  (string loc)
If the length of string is greater than loc, inserts a line feed at the last
space or math symbol before loc, or at loc if no math symbols."
 (cond ((> (length string) loc)
        (let* ((loc1 (find-terminator (substr string 0 loc) loc))
               (s1 (substr string 0 loc1))
               (s2 (substr string loc1 (- (length string) loc1))))
         (format nil "~a~%~a" s1 (split-string s2 :loc loc))))
       (t string)))

(defun find-terminator (string len)
"Function args (string)
Returns the right-most location in string of any of the elements of symbols."
   (let* ((symbols '(#\Space #\+ #\- #\* #\/))
          (n (1- (length string))))
     (cond ((member (select string n) symbols) n)
           ((= n 0) len)
           (t (find-terminator (select string (iseq n)) len)))))
   


;;;
;;; functions to return a permutation or a subsample from (iseq n)
;;;
;;;  For example in dataset d, to redo the analysis on a subset of the data:
;;;  assuming all cases are currently included,
;;;  (send d :toggle-cases (subsample 1000 900))
;;;  will set 900 of the first 1000 cases to not-included

(defun permute (n &optional (size n))
"Function args: (n &optional (size n))
Returns a sample without replacement of size s from (iseq n)"
  (select (order (uniform-rand n)) (iseq size)))

(defun subsample (n s)
"Function args: (n s)
Returns a sample without replacement of size s from (iseq n)"
  (permute n s))

